/*
 * Copyright 2019 by Kappich Systemberatung Aachen
 * ALL RIGHTS RESERVED.
 */

package de.bsvrz.dav.daf.accessControl.internal;

import com.google.common.collect.ImmutableList;
import de.bsvrz.sys.funclib.debug.Debug;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * Klasse, die Basisberechtigungen enthält, die jeder Benutzer implizit zugewiesen bekommt. Es gibt keine Möglichkeit, einem Benutzer über die Parametrierung
 * diese Rechte zu nehmen. Bei Bedarf können weitere Berechtigungen hier hinzugefügt werden. Damit eine Aktion erlaubt ist, müssen Attributgruppe und Aspekt
 * übereinstimmen und die jeweilige Aktion muss über den Boolean-Parameter erlaubt sein. Die Angabe einer leeren Liste für Attributgruppen oder Aspekte bedeutet
 * wie üblich, dass alle Attributgruppen bzw. Aspekte ausgewählt sind.
 *
 * @author Kappich Systemberatung
 * @version $Revision: 0000 $
 */
class ImplicitAccessUnitManager {

	private static final Debug _debug = Debug.getLogger();

	private static final List<AccessUnit> _accessUnits = new ArrayList<>();

	static {
		// Neue Berechtigungen können ab hier eingefügt werden.

		// Basis Anfragen (aus kb.objekteZugriffsrechteBasis)
		_accessUnits.add(
				new AccessUnit(
						// Attributgruppen
						ImmutableList.of(
								"atg.konfigurationsBenutzerverwaltungsAnfrageSchnittstelle",
								"atg.konfigurationsAnfrageSchnittstelleSchreibend",
								"atg.betriebsMeldung",
								"atg.konfigurationsSchreibAnfrage",
								"atg.konfigurationsAnfrage",
								"atg.konfigurationsAnfrageSchnittstelleLesend",
								"atg.konfigurationsBereichsverwaltungsAnfrageSchnittstelle"
						),
						// Aspekte
						ImmutableList.of(
								"asp.anfrage",
								"asp.information"
						),
						true,    // Sender
						false,   // Empfänger
						false,   // Quelle
						false    // Senke
				)
		);
		// Basis Antworten (aus kb.objekteZugriffsrechteBasis)
		_accessUnits.add(
				new AccessUnit(
						// Attributgruppen
						ImmutableList.of(
								"atg.konfigurationsBenutzerverwaltungsAnfrageSchnittstelle",
								"atg.konfigurationsAnfrageSchnittstelleSchreibend",
								"atg.konfigurationsAntwort",
								"atg.konfigurationsAnfrageSchnittstelleLesend",
								"atg.konfigurationsBereichsverwaltungsAnfrageSchnittstelle",
								"atg.konfigurationsSchreibAntwort"
						),
						// Aspekte
						ImmutableList.of(
								"asp.antwort"
						),
						false,   // Sender
						false,   // Empfänger
						false,   // Quelle
						true     // Senke
				)
		);
		// Basis Fertigmeldung (aus kb.objekteZugriffsrechteBasis)
		_accessUnits.add(
				new AccessUnit(
						// Attributgruppen
						ImmutableList.of(
								"atg.applikationsFertigmeldung"
						),
						// Aspekte
						ImmutableList.of(
								"asp.standard"
						),
						false,   // Sender
						true,    // Empfänger
						true,    // Quelle
						false    // Senke
				)
		);
		// Implizite Rechte zum Lesen der Berechtigungen von Benutzern
		_accessUnits.add(
				new AccessUnit(
						// Attributgruppen
						ImmutableList.of(
								"atg.berechtigungsklassen",
								"atg.rollenRegionenPaare",
								"atg.rollenAktivität",
								"atg.region",
								"atg.benutzerParameter",
								"atg.zugriffsRechte",
								"atg.rollenRegionenPaareParameter"
						),
						// Aspekte
						ImmutableList.of(
								"asp.parameterSoll"
						),
						false,   // Sender
						true,    // Empfänger
						false,   // Quelle
						false    // Senke
				)
		);	
		// Implizite Rechte zum Lesen der Rechteprüfungsart
		_accessUnits.add(
				new AccessUnit(
						// Attributgruppen
						ImmutableList.of(
								"atg.datenverteilerRechteprüfung"
						),
						// Aspekte
						ImmutableList.of(
								"asp.standard"
						),
						false,   // Sender
						true,    // Empfänger
						false,   // Quelle
						false    // Senke
				)
		);
	}

	public static boolean isAllowed(final String atg, final String asp, final UserAction action) {
		for(AccessUnit accessUnit : _accessUnits) {
			if(accessUnit.isAllowed(atg,asp,action)) return true;
		}
		return false;
	}

	private static class AccessUnit {

		private final List<String> _attributeGroups;

		private final List<String> _aspects;

		private final boolean _allowSender;

		private final boolean _allowReceiver;

		private final boolean _allowSource;

		private final boolean _allowDrain;

		private AccessUnit(
				final List<String> attributeGroups,
				final List<String> aspects,
				final boolean allowSender,
				final boolean allowReceiver, final boolean allowSource, final boolean allowDrain) {
			_attributeGroups = attributeGroups;
			_aspects = aspects;
			_allowSender = allowSender;
			_allowReceiver = allowReceiver;
			_allowSource = allowSource;
			_allowDrain = allowDrain;
		}

		public boolean isAllowed(final String atg, final String asp, final UserAction action) {
			if(_attributeGroups.size() != 0 && !_attributeGroups.contains(atg)) {
				return false; // Keine Aussage, weil ATG nicht passt.
			}
			if(_aspects.size() != 0 && !_aspects.contains(asp)) {
				return false; // Keine Aussage, weil ASP nicht passt.
			}
			switch(action) {
				case RECEIVER:
					return _allowReceiver;
				case SENDER:
					return _allowSender;
				case SOURCE:
					return _allowSource;
				case DRAIN:
					return _allowDrain;
				default:
					_debug.error("Unbekannte Aktion: " + action);
					return false;
			}
		}
	}
}
